use mirrord_config_derive::MirrordConfig;
use schemars::JsonSchema;

use super::{FsModeConfig, FsUserConfig};
use crate::{
    config::{from_env::FromEnv, source::MirrordConfigSource, ConfigError},
    util::{MirrordToggleableConfig, VecOrSingle},
};

// TODO(alex): We could turn this derive macro (`MirrordConfig`) into an attribute version, which
// would allow us to "capture" the `derive` statement, making it possible to implement the same for
// whatever is generated by `map_to`.
/// Advanced user configuration for file operations.
///
/// Allows the user to specify:
///
/// What is the default behavior for file operations:
/// 1. read - Read from the remote file system (default)
/// 2. write - Read/Write from the remote file system.
/// 3. local - Read from the local file system.
/// 4. disabled - Disable file operations.
///
/// Besides the default behavior, user can specify behavior for specific regex patterns. Case
/// insensitive. 1. read_write - List of patterns that should be read/write remotely.
/// 2. read_only - List of patterns that should be read only remotely.
/// 3. local - List of patterns that should be read locally.
///
/// The logic for choosing the behavior is as follows:
/// 1. Check if one of the patterns match the file path, do the corresponding action. There's no
/// specified order if two lists match the same path, we will use the first one (and we do not
/// guarantee what is first).
/// Note: Specifying the same path in two lists is unsupported and can lead to
///         undefined behaviour.
/// 2. Check our "special list" - we have an internal at compile time list
/// for different behavior based on patterns    to provide better UX.
/// 3. If none of the above match, use the default behavior (mode).
///
/// ## Examples
///
/// - Read `/lib` locally, `/etc` remotely and `/var/run` read write remotely. Rest local
///
/// ```yaml
/// # mirrord-config.yaml
///
/// [fs]
/// mode = read
/// read_write = ["/var/run"]
/// read_only = ["/etc"]
/// local = ["/lib"]
/// ```
#[derive(MirrordConfig, Default, Clone, PartialEq, Eq, Debug)]
#[config(
    map_to = "AdvancedFsUserConfig",
    derive = "PartialEq,Eq,JsonSchema",
    generator = "FsUserConfig"
)]
pub struct FsConfig {
    /// File operations mode, defaults to read-only, see [`FsModeConfig`].
    #[config(nested)]
    pub mode: FsModeConfig,

    /// DEPRECATED
    ///  Allows the user to specify regexes that are used to match against files when mirrord file
    /// operations are enabled.
    ///
    /// The regexes specified here will make mirrord operate only on files that match it, otherwise
    /// the file will be accessed locally (bypassing mirrord).
    #[config(
        env = "MIRRORD_FILE_FILTER_INCLUDE",
        deprecated = "Use access type configuration instead."
    )]
    pub include: Option<VecOrSingle<String>>,

    /// DEPRECATED
    /// Allows the user to specify regexes that are used to match against files when mirrord file
    /// operations are enabled.
    ///
    /// The opposite of `include`, files that match the regexes specified here will bypass mirrord
    /// and are accessed locally.
    #[config(
        env = "MIRRORD_FILE_FILTER_EXCLUDE",
        deprecated = "Use access type configuration instead."
    )]
    pub exclude: Option<VecOrSingle<String>>,

    /// Specify file path patterns that if matched will be read and written to the remote.
    #[config(env = "MIRRORD_FILE_READ_WRITE_PATTERN")]
    pub read_write: Option<VecOrSingle<String>>,

    /// Specify file path patterns that if matched will be read from the remote.
    /// if file matching the pattern is opened for writing or read/write it will be opened locally.
    #[config(env = "MIRRORD_FILE_READ_ONLY_PATTERN")]
    pub read_only: Option<VecOrSingle<String>>,

    /// Specify file path patterns that if matched will be opened locally.
    #[config(env = "MIRRORD_FILE_LOCAL_PATTERN")]
    pub local: Option<VecOrSingle<String>>,
}

impl MirrordToggleableConfig for AdvancedFsUserConfig {
    fn disabled_config() -> Result<Self::Generated, ConfigError> {
        let mode = FsModeConfig::disabled_config()?;
        let include = FromEnv::new("MIRRORD_FILE_FILTER_INCLUDE")
            .source_value()
            .transpose()?;
        let exclude = FromEnv::new("MIRRORD_FILE_FILTER_EXCLUDE")
            .source_value()
            .transpose()?;

        let read_write = FromEnv::new("MIRRORD_FILE_READ_WRITE_PATTERN")
            .source_value()
            .transpose()?;
        let read_only = FromEnv::new("MIRRORD_FILE_READ_ONLY_PATTERN")
            .source_value()
            .transpose()?;
        let local = FromEnv::new("MIRRORD_FILE_LOCAL_PATTERN")
            .source_value()
            .transpose()?;

        Ok(Self::Generated {
            mode,
            include,
            exclude,
            read_write,
            read_only,
            local,
        })
    }
}

impl FsConfig {
    pub fn is_read(&self) -> bool {
        self.mode.is_read()
    }

    pub fn is_write(&self) -> bool {
        self.mode.is_write()
    }

    /// Checks if fs operations are active
    pub fn is_active(&self) -> bool {
        !matches!(self.mode, FsModeConfig::Disabled | FsModeConfig::Local)
    }
}

#[cfg(test)]
mod tests {
    use rstest::rstest;

    use super::*;
    use crate::{config::MirrordConfig, util::testing::with_env_vars};

    #[rstest]
    fn test_advanced_fs_config_default() {
        let expect = FsConfig {
            mode: FsModeConfig::Read,
            ..Default::default()
        };

        with_env_vars(
            vec![
                ("MIRRORD_FILE_OPS", None),
                ("MIRRORD_FILE_RO_OPS", None),
                ("MIRRORD_FILE_FILTER_INCLUDE", None),
                ("MIRRORD_FILE_FILTER_EXCLUDE", None),
            ],
            || {
                let fs_config = AdvancedFsUserConfig::default().generate_config().unwrap();

                assert_eq!(fs_config, expect);
            },
        );
    }

    #[rstest]
    fn test_advanced_fs_config_file_filter_include() {
        let expect = FsConfig {
            mode: FsModeConfig::Read,
            include: Some(VecOrSingle::Single(".*".to_string())),
            ..Default::default()
        };

        with_env_vars(
            vec![
                ("MIRRORD_FILE_OPS", None),
                ("MIRRORD_FILE_RO_OPS", None),
                ("MIRRORD_FILE_FILTER_INCLUDE", None),
                ("MIRRORD_FILE_FILTER_EXCLUDE", None),
            ],
            || {
                let fs_config = AdvancedFsUserConfig {
                    include: Some(VecOrSingle::Single(".*".to_string())),
                    ..Default::default()
                }
                .generate_config()
                .unwrap();

                assert_eq!(fs_config, expect);
            },
        );
    }
}
